/*
 * ProviderSupport.java
 *
 * 作成日: 2008/10/01
 * Copyright(C)2008 Kyushu Institute of Technology., All right reserved. 
 */
package jp.ac.kyutech.SRP.Scripting;

import static java.text.MessageFormat.format;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

import javax.swing.event.EventListenerList;

import jp.go.aist.rtm.RTC.RTObject_impl;
import jp.go.aist.rtm.RTC.port.CorbaPort;

import org.apache.commons.beanutils.MethodUtils;
import org.apache.commons.logging.Log;
import org.omg.PortableServer.Servant;

/**
 * ProviderのRTコンポーネントとインタプリタの協調をサポートします。
 * 
 * TODO: 返り値の扱いを実装する。
 *
 * @author 
 */
public class ProviderSupport {

    /** ロガー */
    protected Log log = org.apache.commons.logging.LogFactory.getLog(ProviderSupport.class);
    
    /**
     * RTコンポーネントを保持します。
     */
    protected RTObject_impl component;
    
    /**
     * Providerのポートを保持します。
     */
    protected CorbaPort port;
    
    /**
     * イベントリスナーを保持します。
     */
    protected EventListenerList listenerList;

    /**
     * デフォルトコンストラクタ。
     */
    public ProviderSupport() {
        this(null);
    }

    /**
     * サーバントを指定するコンストラクタ。
     * 
     * @param servant
     */
    public ProviderSupport(Servant servant) {
        super();
        this.listenerList = new EventListenerList();
    }

    /**
     * RTコンポーネントとProvicerのポートを指定するコンストラクタ。
     * 
     * @param component
     * @param port
     */
    public ProviderSupport(RTObject_impl component, CorbaPort port) {
        super();
        this.component = component;
        this.port = port;
        this.listenerList = new EventListenerList();
    }
    
    /**
     * @return component を戻します。
     */
    public RTObject_impl getComponent() {
        return component;
    }

    /**
     * @param component 設定する component
     */
    public void setComponent(RTObject_impl component) {
        this.component = component;
    }

    /**
     * @return port を戻します。
     */
    public CorbaPort getPort() {
        return port;
    }

    /**
     * @param port 設定する port
     */
    public void setPort(CorbaPort port) {
        this.port = port;
    }

    /**
     * @return listenerList を戻します。
     */
    public EventListenerList getListenerList() {
        return listenerList;
    }

    /**
     * @param listenerList 設定する listenerList
     */
    public void setListenerList(EventListenerList listenerList) {
        this.listenerList = listenerList;
    }
    
    /**
     * RTコンポーネント用のイベントリスナを追加します。
     * 
     * @param providerListener 追加するリスナ
     */
    public void addProviderListener(ProviderListener providerListener) {
        this.listenerList.add(ProviderListener.class, providerListener);
    }
    
    /**
     * RTコンポーネント用のイベントリスナを削除します。
     * 
     * @param providerListener
     */
    public void removeProviderListener(ProviderListener providerListener) {
        this.listenerList.remove(ProviderListener.class, providerListener);
    }
    
    /**
     * RTコンポーネント用のイベントリスナを全て削除します。
     */
    public void clearProviderListener() {
        Object[] listenerList = this.listenerList.getListenerList();
        if (listenerList != null) {
            for (Object listener: listenerList) {
                if (listener instanceof ProviderListener) {
                    this.listenerList.remove(ProviderListener.class, (ProviderListener )listener);
                }
            }
        }
    }
    
    /**
     * メソッド名と引数を指定して、サーバントメソッドの呼び出しイベントを通知します。
     * 
     * @param methodName
     * @param args
     */
    public Object fireProviderInvoked(String methodName, Object[] args) {
        ProviderEvent e = new ProviderEvent(this);
        e.setMethodName(methodName);
        e.setArguments(args);
        return fireProviderInvoked(e);
    }
    
    /**
     * サーバントメソッドの呼び出しイベントを通知します。
     * 
     * @param e
     */
    public Object fireProviderInvoked(ProviderEvent e) {
        boolean firstListener = true;
        Object ret = null;
        if (e == null) {
            throw new IllegalArgumentException("ProvicerEvent is null");
        }
        for (ProviderListener l : this.listenerList.getListeners(ProviderListener.class)) {
            if (l != null) {
                boolean foundBshMethod = false;
                // リスナが Proxy インスタンスの場合は、BeanShell のブロックを呼び出す。
                if (Proxy.isProxyClass(l.getClass())) {
                    InvocationHandler handler = Proxy.getInvocationHandler(l);
                    // bsh.XThis$Handler#invokeMethod(String, Object[]) を呼び出す。
                    // この処理により、BeanShellのスクリプト上のメソッドが呼び出される
                    Method method = MethodUtils.getMatchingAccessibleMethod(handler.getClass(), "invokeMethod", new Class[] {String.class, Object[].class});
                    if (method != null) {
                        foundBshMethod = true;
                        try {
                            Object curRet = MethodUtils.invokeMethod(handler, "invokeMethod", new Object[] {e.getMethodName(), e.getArguments()});
                            if (firstListener) {
                                ret = curRet;
                            }
                            firstListener = false;
                        }
                        //catch (NoSuchMethodException ex) {
                        //    throw new RuntimeException(ex);
                        //}
                        //catch (IllegalAccessException ex) {
                        //    throw new RuntimeException(ex);
                        //}
                        catch (InvocationTargetException ex) {
                            // NOTE: ProviderListener が BeanShell の AnonymousBlock(動的に定義されたブロック) の場合で、
                            //   呼び出すメソッドが見つからない場合は、この例外が投げられる。
                            String message = format("Fail to invoke proxy method: listener={0}, event={1}" +
                                    ", !!!CHECK!!! existence of provider method.", l.toString(), e.toString());
                            log.error(message, ex);
                            throw new org.apache.commons.lang.exception.NestableRuntimeException(message, ex);
                        }
                        catch (Throwable ex) {
                            String message = format("Fail to invoke proxy method: listener={0}, event={1}", l.toString(), e.toString());
                            log.error(message, ex);
                            throw new org.apache.commons.lang.exception.NestableRuntimeException(message, ex);
                        }
                    }
                }
                // リスナが BeanShell のブロックでない場合の処理
                if (foundBshMethod == false) {
                    try {
                        Object curRet = l.invoked(e);
                        if (firstListener) {
                            ret = curRet;
                        }
                        firstListener = false;
                    }
                    catch (Throwable ex) {
                        String message = format("Fail to invoke method: listener={0}, event={1}", l.toString(), e.toString());
                        log.error(message, ex);
                        throw new org.apache.commons.lang.exception.NestableRuntimeException(message, ex);
                    }
                }
            }
        }
        return ret;
    }
    
}
